/*
    UKI_Control UKI HID Control application and example code for interface
    Copyright (C) 2010  NOP Security

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <hid.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pwd.h>
#include "USBKeyInject.h"
#include "scancode_map.h"

#define	TRUE						1
#define	FALSE						0

#define	EEMEMSIZE				0x200

#define	HIDCTL_IDLE				0
#define	HIDCTL_ENABLE			1
#define	HIDCTL_DISABLE			2
#define	HIDCTL_BLINK			3
#define	HIDCTL_DUMP				4
#define	HIDCTL_UPLD				5
#define	HIDCTL_DNLD				6
#define	HIDCTL_SPGE				7
#define	HIDCTL_GPGE				8
#define	HIDCTL_ECHO				9
#define	HIDCTL_STADDR			10
#define	HIDCTL_END				-1
#define	HIDCTL_LIST_DATA		-2

typedef struct UKIScancodeBuf {
	unsigned char code;
	unsigned char modifier;
} UKIScancode_t, *pUKIScancode_t;

unsigned char const PLEN=2; int const PINP[]={0x00010000, 0x00010000};
char intcmd_help[]="\
Command list:\n\
enable - enables device output\n\
disable - disables device output\n\
dump - displays (length) bytes of device EEPROM from (address) location\n\
upload - uploads (length) bytes of device EEPROM to (address) location from (filename) file\n\
download - uploads (length) bytes of device EEPROM from (address) location to (filename) file\n\
setpage # - sets current EEPROM page to #\n\
getpage - gets current EEPROM page\n\
startaddr # - sets current starting address in external EEPROM to address # on page 0, or -1 for base EEPROM\n\
echo - echos characters via UKI\n\
address # - sets (address) to #\n\
length # - sets (length) to #\n\
filename *** - sets (filename) to ***\n\
quit - ends session\n\
\n";

bool matcher(struct usb_dev_handle *usbdev, void *chkstr, unsigned int bsize)
{char serbuf[257];
	
	memset(serbuf, 0, 257);
	usb_get_string_simple(usbdev, usb_device(usbdev)->descriptor.iSerialNumber, serbuf, 256);
	if(!strncmp(serbuf, (char *)chkstr, bsize)) return(TRUE);
	return(FALSE);
}

int UKI_Control_Cmd(HIDInterface *hidtrg, unsigned char cmd, unsigned char argA, unsigned char argB, unsigned char argC)
{unsigned char databuf[5]; int ctmout=0; hid_return retval;

	databuf[0]=2;
	databuf[1]=cmd;
	databuf[2]=argA;
	databuf[3]=argB;
	databuf[4]=argC;
	
	retval=hid_set_output_report(hidtrg, PINP, PLEN, (char *)databuf, 5);
	if(retval!=HID_RET_SUCCESS) return(-1);
	
	do{ retval=hid_get_input_report(hidtrg, PINP, PLEN, (char *)databuf, 3);
	} while(retval==HID_RET_SUCCESS && databuf[1] != STATUSR_COMPLETE && databuf[1] != STATUSR_EOF && ctmout++<20);
	
	if(retval!=HID_RET_SUCCESS || databuf[1] == STATUSR_EOF) return(-1);
	else if(databuf[1] == STATUSR_COMPLETE) return(databuf[2]);

	return(-2);
}

void list_commands()
{
	write(0, intcmd_help, sizeof(intcmd_help));
	return;
}

void usage(char *execnm)
{
	printf("\nUsage: %s [-v vendor ID (default:16C0)] [-p product ID (default:27DB)] [-s serial (default:'NOPSecurity.net:UKI') ] [-c command] [-a address] [-l length] [-f filename] [-e EEPROM page]\n\n", execnm);
	printf("\
UKI Controller\n\
(c) 2010 NOP Security\n\n\
Commands include:\n\
\tEnable - Enable UKI run mode\n\
\tDisable - Disable UKI run mode\n\
\tBlink - Blink test LED\n\
\tDump - Dump UKI memory contents to screen\n\
\tUpload - Upload from <filename> into EEPROM\n\
\tDownload - Download EEPROM into <filename> file\n\
\tSetPage - Set EEPROM page to <EEPROM page>: 0 for external (if equipped) or -1 for internal EEPROM\n\
\tGetPage - Get current EEPROM page\n\
\tStartAddr - sets current starting address in external EEPROM: <address> for external address (if equipped) or -1 for internal EEPROM\n\
\tEcho - Echo characters using UKI\n\n\
<address> is the address of the EEPROM to begin at. (default:0)\n\
<length> is the data length of the EEPROM or file operation to perform.\n\n\
-i enables interactive mode.\n\n");
	return;
}

int cvtToCtlCodes(unsigned char *ascii_buf, int ascii_sz, pUKIScancode_t ctlcode_buf, int ctlcode_sz, unsigned char modifiers)
{unsigned char lastbyte=-1; pUKIScancode_t curctlc=ctlcode_buf; int x;
	
	for(x=0; x<ascii_sz && (curctlc+1)<(ctlcode_buf+ctlcode_sz); x++){
		if(ascii_buf[x] == '~' && ascii_buf[++x] != '~' && ascii_buf[x] >= 0x30 && ascii_buf[x] <= 0x7A){
			if(lastbyte == AsciiToSpecialMap[ascii_buf[x]-0x30]){
				curctlc->code     = 0;
				curctlc->modifier = modifiers;
				curctlc++;
				lastbyte=0;
				if((curctlc+1)>=(ctlcode_buf+ctlcode_sz)) return(curctlc-ctlcode_buf);
			}
			curctlc->code     = AsciiToSpecialMap[ascii_buf[x]-0x30];
			curctlc->modifier = modifiers;
			curctlc++;
			lastbyte = AsciiToSpecialMap[ascii_buf[x]-0x30];
		}
		else {
			if(lastbyte == AsciiToScancodeMap[(ascii_buf[x]<<1)+1]){
				curctlc->code     = 0;
				curctlc->modifier = modifiers;
				curctlc++;
				lastbyte=0;
				if((curctlc+1)>=(ctlcode_buf+ctlcode_sz)) return(curctlc-ctlcode_buf);
			}
			curctlc->code     = AsciiToScancodeMap[(ascii_buf[x]<<1)+1];
			curctlc->modifier = AsciiToScancodeMap[(ascii_buf[x]<<1)] ^ modifiers;
			curctlc++;
			lastbyte = AsciiToScancodeMap[(ascii_buf[x]<<1)+1];
		}
	}
	if(lastbyte != 0){
		if((curctlc+1)>=(ctlcode_buf+ctlcode_sz)) return(curctlc-ctlcode_buf);
		curctlc->code     = 0;
		curctlc->modifier = 0;
	}
	return(curctlc-ctlcode_buf);
}

int main(int argc, char **argv)
{int optionr, vendid=0x16C0, prodid=0x27DB, ccmd=HIDCTL_IDLE, interac=0, drpvusr=0, drpvgrp=0, curpge=-1; char cmdbuf[512], *cmddta=cmdbuf;
 HIDInterface *hidtrg; HIDInterfaceMatcher hidmatch; hid_return retval; unsigned char databuf[8], *transbuf;
 int cmdaddr=0, cmdlen=-1, tmpx, tmpy, ctmout, datafd, echct, staddr=0, doffst;
 char datafilename[512], serialnm[257]="NOPSecurity.net:UKI", echobuf[32]; UKIScancode_t scbuf[64];
 struct stat dfstat;
 struct passwd *upwdta;

	if(argc == 1){usage(argv[0]);return(0);}
	
	while((optionr=getopt(argc, argv, "hLiu:v:p:s:c:a:l:f:e")) != -1){
		switch(optionr){
			case 'v':	if(optarg) vendid=strtol(optarg, NULL, 16);
				break;
			case 'p':	if(optarg) prodid=strtol(optarg, NULL, 16);
				break;
			case 's':	if(optarg) strncpy(serialnm, optarg, 256);
				break;
			case 'L':	ccmd=HIDCTL_LIST_DATA;
				break;
			case 'e':	if(optarg) curpge=strtol(optarg, NULL, 0);
				break;
			case 'c':	if(!optarg) break;
							if(strncasecmp(optarg, "Enable", 8)==0) ccmd=HIDCTL_ENABLE;
							if(strncasecmp(optarg, "Disable", 8)==0) ccmd=HIDCTL_DISABLE;
							if(strncasecmp(optarg, "Blink", 8)==0) ccmd=HIDCTL_BLINK;
							if(strncasecmp(optarg, "Dump", 8)==0) ccmd=HIDCTL_DUMP;
							if(strncasecmp(optarg, "Upload", 8)==0) ccmd=HIDCTL_UPLD;
							if(strncasecmp(optarg, "Download", 8)==0) ccmd=HIDCTL_DNLD;
							if(strncasecmp(optarg, "SetPage", 8)==0) ccmd=HIDCTL_SPGE;
							if(strncasecmp(optarg, "GetPage", 8)==0) ccmd=HIDCTL_GPGE;
							if(strncasecmp(optarg, "StartAddr", 10)==0) ccmd=HIDCTL_STADDR;
							if(strncasecmp(optarg, "Echo", 8)==0) ccmd=HIDCTL_ECHO;
				break;
			case 'u':	if(!optarg) break;
							upwdta=getpwnam(optarg);
							if(!upwdta) break;
							drpvusr=upwdta->pw_uid;
							drpvgrp=upwdta->pw_gid;
				break;
			case 'a':	if(optarg) cmdaddr=strtol(optarg, NULL, 0);
							staddr=cmdaddr;
							if(cmdaddr<0||cmdaddr>0x10000) cmdaddr=0;
							if(ccmd==HIDCTL_STADDR) printf("Start address set to 0x%04X\n", staddr);
							printf("Address set to 0x%04X\n", cmdaddr);
				break;
			case 'l':	if(optarg) cmdlen=strtol(optarg, NULL, 0);
							if(cmdlen<0||cmdlen>0x10000) cmdlen=0;
							printf("Length set to 0x%04X\n", cmdlen);
				break;
			case 'f':	if(optarg) strncpy(datafilename, optarg, 511);
				break;
			case 'i':	interac=1;
				break;
			default:		usage(argv[0]); return(0);
				break;
		}
	}

	if(vendid==-1||prodid==-1){
		printf("Enter the vendor ID and the product ID for the device.\n");
		return(-1);
	}

	if(!getuid() && !drpvusr){
		printf("Running as root. Please enter the user to drop privs to with -u.\n");
		return(-1);
	}
	
	hidmatch.vendor_id=vendid;
	hidmatch.product_id=prodid;
	hidmatch.matcher_fn=(void *)matcher;
	hidmatch.custom_data=(void *)serialnm;
	hidmatch.custom_data_length=strlen(serialnm);
	
	if((retval=hid_init())!=HID_RET_SUCCESS){
		printf("HID Init failed: %d\n", retval);
		return(-1);
	}
	if((hidtrg=hid_new_HIDInterface())==0){
		printf("Interface allocation failed.\n");
		return(-1);
	}
	if((retval=hid_force_open(hidtrg, 0, &hidmatch, 3))!=HID_RET_SUCCESS){
		printf("HID Open failed: %d\n", retval);
		return(-1);
	}

	
	if(ccmd == HIDCTL_LIST_DATA){
		printf("Searching for %04X:%04X\n", vendid, prodid);
		if((retval=hid_write_identification(stdout, hidtrg))!=HID_RET_SUCCESS){
			printf("HID ident output failed: %d\n", retval);
			return(-1);
		}
	
		if((retval=hid_dump_tree(stdout, hidtrg))!=HID_RET_SUCCESS){
			printf("HID tree output failed: %d\n", retval);
			return(-1);
		}
		ccmd=HIDCTL_END;
	}
	
	memset(datafilename, 0, 512);
	setgid(drpvgrp);
	setuid(drpvusr);
	
	while(ccmd != HIDCTL_END){
		
		databuf[0]=0x02;
		switch(ccmd){
			case HIDCTL_BLINK:	databuf[1]=CONTROL_BLINK;
										databuf[2]=0x05;
				break;
			case HIDCTL_ENABLE:	databuf[1]=CONTROL_ENABLE;
				break;
			case HIDCTL_DISABLE:	databuf[1]=CONTROL_DISABLE;
				break;
			case HIDCTL_DUMP:		for(tmpx=0; tmpx<cmdlen; tmpx++){
											databuf[1]=CONTROL_READ_EEBLOC;
											databuf[2]=(cmdaddr+tmpx)&0xFF;
											databuf[3]=((cmdaddr+tmpx)>>8)&0xFF;
											
											retval=hid_set_output_report(hidtrg, PINP, PLEN, (char *)databuf, 5);
											if(retval!=HID_RET_SUCCESS) break;
											
											ctmout=0;
											do{ retval=hid_get_input_report(hidtrg, PINP, PLEN, (char *)databuf, 3);
											} while(retval==HID_RET_SUCCESS && databuf[1] != STATUSR_COMPLETE && databuf[1] != STATUSR_EOF && ctmout++<20);
											
											if((tmpx&0xF)==0) printf("\n%04X: ", tmpx&0xFFF0);
											if(databuf[1] == STATUSR_COMPLETE) printf("%02X:", databuf[2]);
											else if(databuf[1] != STATUSR_EOF) printf("--:");
											else tmpx=cmdlen;
										}
										
										printf("\n");
				break;
			case HIDCTL_UPLD:		if(!datafilename[0]) break;
										datafd=open(datafilename, O_RDONLY);
										if(datafd==-1){perror("open"); ccmd=HIDCTL_END; break;}
										if(fstat(datafd, &dfstat)==-1){perror("fstat"); ccmd=HIDCTL_END; close(datafd); break;}
										if((transbuf=(unsigned char *)mmap(NULL, dfstat.st_size, PROT_READ, MAP_PRIVATE, datafd, 0))==MAP_FAILED){
											perror("mmap"); ccmd=HIDCTL_END; close(datafd); break;
										}
										curpge=UKI_Control_Cmd(hidtrg, CONTROL_GET_EXEEPAGE, 0, 0, 0);
										if(curpge==-1) doffst=6;
										else doffst=0;
										for(tmpx=0; tmpx<cmdlen && tmpx<dfstat.st_size; tmpx++){
											databuf[1]=CONTROL_WRITE_EEBLOC;
											databuf[2]=(cmdaddr+doffst+tmpx)&0xFF;
											databuf[3]=((cmdaddr+doffst+tmpx)>>8)&0xFF;
											databuf[4]=transbuf[tmpx];
											
											retval=hid_set_output_report(hidtrg, PINP, PLEN, (char *)databuf, 5);
											if(retval!=HID_RET_SUCCESS) break;
											
											ctmout=0;
											do{ retval=hid_get_input_report(hidtrg, PINP, PLEN, (char *)databuf, 3);
											} while(retval==HID_RET_SUCCESS && databuf[1] != STATUSR_COMPLETE && databuf[1] != STATUSR_EOF && ctmout++<20);
											
											if(databuf[1] != STATUSR_COMPLETE) break;

											if(tmpx && (tmpx&0x3FF)==0) printf("Uploaded %d bytes.\n", tmpx);
										}
										
										munmap(transbuf, dfstat.st_size);
										close(datafd);
										printf("Uploaded %d bytes.\n", tmpx);
				break;
			case HIDCTL_DNLD:		if(!datafilename[0]) break;
										datafd=open(datafilename, O_RDWR|O_CREAT|O_TRUNC, 0644);
										if(datafd==-1){perror("creat"); ccmd=HIDCTL_END; break;}
										curpge=UKI_Control_Cmd(hidtrg, CONTROL_GET_EXEEPAGE, 0, 0, 0);
										if(curpge==-1) doffst=6;
										else doffst=0;
										for(tmpx=0; tmpx<cmdlen; tmpx++){
											databuf[1]=CONTROL_READ_EEBLOC;
											databuf[2]=(cmdaddr+doffst+tmpx)&0xFF;
											databuf[3]=((cmdaddr+doffst+tmpx)>>8)&0xFF;
											
											retval=hid_set_output_report(hidtrg, PINP, PLEN, (char *)databuf, 5);
											if(retval!=HID_RET_SUCCESS) break;
											
											ctmout=0;
											do{ retval=hid_get_input_report(hidtrg, PINP, PLEN, (char *)databuf, 3);
											} while(retval==HID_RET_SUCCESS && databuf[1] != STATUSR_COMPLETE && databuf[1] != STATUSR_EOF && ctmout++<20);
											
											if(retval!=HID_RET_SUCCESS){printf("input report error\n"); break;}
											if(databuf[1] == STATUSR_COMPLETE) write(datafd, databuf+2, 1);
											else break;
											
											if(tmpx && (tmpx&0x3FF)==0) printf("Downloaded and saved %d bytes.\n", tmpx);
										}
										
										close(datafd);
										printf("Downloaded and saved %d bytes.\n", tmpx);
				break;
			case HIDCTL_GPGE:		curpge=UKI_Control_Cmd(hidtrg, CONTROL_GET_EXEEPAGE, 0, 0, 0);
										printf("Current EEPROM is: %s.\n", curpge>=0?"external":"internal");
				break;
			case HIDCTL_SPGE:	
										retval=UKI_Control_Cmd(hidtrg, CONTROL_SET_EXEEPAGE, curpge&0xFF, (curpge>>8)&0xFF, 0);
										if(retval==-1){ printf("Failed to set EEPROM page.\n"); break;}
										curpge=UKI_Control_Cmd(hidtrg, CONTROL_GET_EXEEPAGE, 0, 0, 0);
										printf("Current EEPROM is: %s.\n", curpge>=0?"external":"internal");
				break;
			case HIDCTL_ECHO:		printf("Echoing %d characters.\n", echct);
										for(tmpx=0; tmpx<echct; tmpx++){
											tmpy=UKI_Control_Cmd(hidtrg, CONTROL_QECHOCHAR, scbuf[tmpx].code, scbuf[tmpx].modifier, 0);
											if(tmpy<0){ printf("Error in character echo @ %d.\n", tmpx); tmpx=echct;}
										}
										echct=0;
				break;
			case HIDCTL_STADDR:	
										retval=UKI_Control_Cmd(hidtrg, CONTROL_SET_EXEEPAGE, 0xFF, 0xFF, 0);
										if(retval==-1){ printf("Failed to set EEPROM page.\n"); break;}

										retval=UKI_Control_Cmd(hidtrg, CONTROL_WRITE_EEBLOC, 0, 0, staddr&0xFF);
										if(retval==-1){ printf("Failed to write EEPROM.\n"); break;}
										retval=UKI_Control_Cmd(hidtrg, CONTROL_WRITE_EEBLOC, 1, 0, (staddr>>8)&0xFF);
										if(retval==-1){ printf("Failed to write EEPROM.\n"); break;}

										if(staddr!=-1) retval=UKI_Control_Cmd(hidtrg, CONTROL_WRITE_EEBLOC, 2, 0, 0);
										else retval=UKI_Control_Cmd(hidtrg, CONTROL_WRITE_EEBLOC, 2, 0, 0xFF);
										if(retval==-1){ printf("Failed to write EEPROM.\n"); break;}
										printf("Start address set to %08X\n", staddr);
										
										retval=UKI_Control_Cmd(hidtrg, CONTROL_SET_EXEEPAGE, curpge&0xFF, (curpge>>8)&0xFF, 0);
										if(retval==-1){ printf("Failed to set EEPROM page.\n"); break;}
										curpge=UKI_Control_Cmd(hidtrg, CONTROL_GET_EXEEPAGE, 0, 0, 0);
										printf("Current EEPROM is: %s.\n", curpge>=0?"external":"internal");
				break;
			case HIDCTL_LIST_DATA:	if((retval=hid_write_identification(stdout, hidtrg))!=HID_RET_SUCCESS){
												printf("HID ident output failed: %d\n", retval); return(-1);
											}
											if((retval=hid_dump_tree(stdout, hidtrg))!=HID_RET_SUCCESS){
												printf("HID tree output failed: %d\n", retval); return(-1);
											}
											ccmd=HIDCTL_END;
				break;
			default:	databuf[1]=0;
				break;
		}

		if(databuf[1] && (retval=hid_set_output_report(hidtrg, PINP, PLEN, (char *)databuf, 5))!=HID_RET_SUCCESS){
			printf("HID output failed: %d\n", retval);
			return(-1);
		}
		
		if((retval=hid_get_input_report(hidtrg, PINP, PLEN, (char *)databuf, 3))!=HID_RET_SUCCESS){
			printf("HID input failed: %d\n", retval);
			return(-1);
		}
		else{
			printf("Response: %02X, %02X, %02X\n", databuf[0], databuf[1], databuf[2]);
			switch(databuf[1]){
				case STATUSR_IDLE:
					printf("Status: UKI Idle\n");
					break;
				case STATUSR_COMPLETE:
					printf("Status: Command Complete\n");
					break;
				case STATUSR_RETRY:
					printf("Status: Retry last command\n");
					break;
				case STATUSR_EOF:
					printf("Status: EOF Reached\n");
					break;
				case STATUSR_EEBUSY:
					printf("Status: EEPROM Busy\n");
					break;
				case STATUSR_BUSY:
					printf("Status: Busy\n");
					break;
				default:
					break;
			}
		}
		
		databuf[1]=0;
		ccmd=HIDCTL_END;
		
		while(interac && ccmd==HIDCTL_END){
			memset(cmdbuf, 0, 512); cmddta=cmdbuf;
			if(cmdlen<0) printf("%04X:~> ", cmdaddr);
			else printf("%04X:%04X> ", cmdaddr, cmdlen);
			fgets(cmdbuf, 500, stdin); cmddta=strsep(&cmddta, "\n\r");
			
			if(!strcasecmp(cmddta, "quit")||!strcasecmp(cmddta, "exit")||!strcasecmp(cmddta, "q")||!strcasecmp(cmddta, "x")){ ccmd=HIDCTL_END; interac=0;}
			else if(!strcasecmp(cmddta, "enable")) ccmd=HIDCTL_ENABLE;
			else if(!strcasecmp(cmddta, "disable")) ccmd=HIDCTL_DISABLE;
			else if(!strcasecmp(cmddta, "dump")) ccmd=HIDCTL_DUMP;
			else if(!strcasecmp(cmddta, "list")) ccmd=HIDCTL_LIST_DATA;
			else if(!strcasecmp(cmddta, "download")) ccmd=HIDCTL_DNLD;
			else if(!strcasecmp(cmddta, "upload")) ccmd=HIDCTL_UPLD;
			else if(!strcasecmp(cmddta, "blink")) ccmd=HIDCTL_BLINK;
			else if(!strcasecmp(cmddta, "getpage")) ccmd=HIDCTL_GPGE;
			else if(!strcasecmp(cmddta, "ls")) system("ls");
			else if(!strncmp(cmddta, "setpage ", sizeof("length ")-1)){
				strsep(&cmddta, " "); curpge=strtol(cmddta, NULL, 0);
				ccmd=HIDCTL_SPGE;
			}
			else if(!strncmp(cmddta, "length ", sizeof("length ")-1)){
				strsep(&cmddta, " "); cmdlen=strtol(cmddta, NULL, 0);
				if(cmdlen<0||cmdlen>0x10000) cmdlen=0;
				printf("Length set to 0x%04X\n", cmdlen);
			}
			else if(!strncmp(cmddta, "address ", sizeof("address ")-1)){
				strsep(&cmddta, " "); cmdaddr=strtol(cmddta, NULL, 0);
				if(cmdaddr<0||cmdaddr>0x10000) cmdaddr=0;
				printf("Address set to 0x%04X\n", cmdaddr);
			}
			else if(!strncmp(cmddta, "startaddr ", sizeof("length ")-1)){
				strsep(&cmddta, " "); staddr=strtol(cmddta, NULL, 0);
				ccmd=HIDCTL_STADDR;
			}
			else if(!strncmp(cmddta, "filename ", sizeof("filename ")-1)){
				strsep(&cmddta, " ");
				memset(datafilename, 0, 512); strncpy(datafilename, cmddta, 511);
				printf("Filename set to %s\n", datafilename);
			}
			else if(!strncmp(cmddta, "echo ", sizeof("echo ")-1)){
				strsep(&cmddta, " ");
				memset(echobuf, 0, 32); strncpy(echobuf, cmddta, 31);
				echct=cvtToCtlCodes( (unsigned char *) echobuf, strlen(echobuf), scbuf, 63, 0);
				ccmd=HIDCTL_ECHO;
			}
			else if(!strcasecmp(cmddta, "help")||!strcasecmp(cmddta, "?")) list_commands();
			else system(cmddta);
			
		}
	}

	if((retval=hid_close(hidtrg))!=HID_RET_SUCCESS){
		printf("HID close failed: %d\n", retval);
		return(-1);
	}
	hid_delete_HIDInterface(&hidtrg);
	if((retval=hid_cleanup())!=HID_RET_SUCCESS){
		printf("HID cleanup failed: %d\n", retval);
		return(-1);
	}
	
	return(0);
}




