/*
    ns_nc_prompt curses code for interactive terminal
    Copyright (C) 2010  NOP Security

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdarg.h>
#include <curses.h>
#include <string.h>

#define MAXSBLS	0x20
typedef struct SBLIST {
	struct SBLIST *next, *prev;
	char	*buffer;
	int	size;
} sbl_obj;
sbl_obj sblhead, *cursbl;
int sbl_ct;
bool curses_on;
int cmdpos, cmdmax;
WINDOW *opwin;

sbl_obj * add_sb_after(sbl_obj *sbltgt, int bufsize)
{sbl_obj *temp;
	if(sbl_ct<MAXSBLS){
		temp = (sbl_obj *) malloc(sizeof(sbl_obj));
		if(!temp) return(NULL);
		memset(temp, 0, sizeof(sbl_obj));
		temp->prev=sbltgt;
		temp->next=sbltgt->next;
		sbltgt->next=temp;
		temp->next->prev=temp;
		sbl_ct++;
		if(bufsize){
			temp->buffer=(char *)malloc(bufsize);
			if(temp->buffer){ memset(temp->buffer, 0, bufsize); temp->size=bufsize;}
		}
	}
	else{
		temp=sbltgt->next;
		if(temp->size>=bufsize) return(temp);
		if(temp->buffer){ free(temp->buffer); temp->buffer=NULL; temp->size=0;}
		if(bufsize){
			temp->buffer=(char *)malloc(bufsize);
			if(temp->buffer){ memset(temp->buffer, 0, bufsize); temp->size=bufsize;}
		}
	}
	return(temp);
}

sbl_obj * del_sb(sbl_obj *sbltgt)
{sbl_obj *temp;
	if(sbl_ct>1){
		temp=sbltgt;
		if(temp->buffer){ free(temp->buffer); temp->buffer=NULL; temp->size=0;}
		temp->prev->next=temp->next;
		temp->next->prev=temp->prev;
		sbl_ct--;
		temp=temp->next;
		free(sbltgt);
	}
	else temp=sbltgt;
	return(temp);
}

sbl_obj * srch_sb_r(sbl_obj *sbltgt, char *tstr, int strsz)
{sbl_obj *temp;
	temp=sbltgt->prev;
	do{
		if(temp->buffer && strsz <= temp->size && !memcmp(temp->buffer, tstr, strsz)) return(temp);
		temp=temp->prev;
	} while(temp != sbltgt);
	return(NULL);
}

int parse_cmdln(char *cmdln, int cmdln_n)
{int cmdbufret=0, cx, cy, chbuf, icmode=1, contcap=1;
	if(!curses_on){
		fgets(cmdln, cmdln_n, stdin);
		cmdln = strsep(&cmdln, "\n");
		return(strlen(cmdln));
	}
	getyx(stdscr, cy, cx);
	if(cy != cmdpos){ move(cmdpos, 0); cmdbufret=0; clrtoeol();}
	cursbl=&sblhead;

	memset(cmdln, 0, cmdln_n);
	while(contcap){
		chbuf=getch();
		getyx(stdscr, cy, cx);
		switch(chbuf){
			case '\t':
				mvinnstr(cmdpos, 0, cmdln, cx);
				cursbl=srch_sb_r(cursbl, cmdln, cx);
				if(cursbl){ move(cmdpos, 0); clrtoeol(); addnstr(cursbl->buffer, cursbl->size);}
				else cursbl=&sblhead;
				memset(cmdln, 0, cmdln_n); move(cmdpos, cx);
				break;
			case 27:
				move(cmdpos, 0); clrtoeol();
				break;
			case KEY_UP:
				move(cmdpos, 0); clrtoeol();
				if(cursbl->prev != &sblhead) cursbl=cursbl->prev;
				if(cursbl->buffer) addnstr(cursbl->buffer, cursbl->size);
				break;
			case KEY_DOWN:
				move(cmdpos, 0); clrtoeol();
				if(cursbl != &sblhead) cursbl=cursbl->next;
				if(cursbl->buffer) addnstr(cursbl->buffer, cursbl->size);
				break;
			case KEY_PPAGE:
				move(cmdpos, 0); clrtoeol();
				cursbl=sblhead.next;
				if(cursbl->buffer) addnstr(cursbl->buffer, cursbl->size);
				break;
			case KEY_NPAGE:
				move(cmdpos, 0); clrtoeol();
				cursbl=&sblhead;
				break;
			case KEY_LEFT:
				if(cx) move(cmdpos, cx-1);
				break;
			case KEY_RIGHT:
				if(cx<cmdmax) move(cmdpos, cx+1);
				break;
			case KEY_BACKSPACE:
				if(cx){ move(cmdpos, cx-1); delch();}
				break;
			case KEY_DC:
				delch();
				break;
			case KEY_IC:
				icmode^=1;
				break;
			case KEY_EIC:
				icmode=1;
				break;
			case KEY_ENTER:
			case '\r':
			case '\n':
				contcap=0;
				break;
			default:
				if(cx < cmdmax-1){ if(icmode) insch(' '); echochar(chbuf); cmdln[cx]=chbuf&0xFF;}
				break;
		}
	}
	
	cmdbufret=mvinnstr(cmdpos, 0, cmdln, cmdln_n);
	if(cmdbufret>=cmdmax) cmdbufret=cmdmax-1; cmdln[cmdbufret]=0;
	if(cmdbufret){
		cmdbufret--;
		while(cmdbufret && cmdln[cmdbufret]==' ') cmdln[cmdbufret--]=0;
		if(cmdln[cmdbufret]!=' ') cmdbufret++;
	}
	if(cmdbufret && !(sblhead.prev->buffer && sblhead.prev->size==cmdbufret && !memcmp(sblhead.prev->buffer, cmdln, cmdbufret))){
		cursbl=add_sb_after(sblhead.prev, cmdbufret);
		if(cursbl && cursbl->buffer) memcpy(cursbl->buffer, cmdln, cmdbufret);
		else cursbl=&sblhead;
	}
	move(cmdpos, 0); clrtoeol();
	return(cmdbufret);
}

void initdisp(bool curseon)
{
	curses_on=curseon;
	if(!curses_on) return;
	initscr(); cbreak(); keypad(stdscr, TRUE); nonl();	noecho(); erase();
	cmdpos=LINES-2;
	cmdmax=COLS;
	opwin=newwin(LINES-4, 0, 0, 0); scrollok(opwin, TRUE);
	memset(&sblhead, 0, sizeof(sblhead)); sbl_ct=1; cursbl=&sblhead; sblhead.next=&sblhead; sblhead.prev=&sblhead;
	refresh();
	return;
}

void stopdisp()
{
	if(!curses_on) return;
	delwin(opwin);
	endwin();
	return;
}

int addOpTextf(char *format, ...)
{char chrbuf[2048]; int tmpln; va_list lptr;
	va_start(lptr, format);
	tmpln=vsnprintf(chrbuf, 2048, format, lptr);
	va_end(lptr);
	if(!curses_on){
		printf("%s", chrbuf);
		return(1);
	}
	waddnstr(opwin, chrbuf, tmpln);
	if(strchr(chrbuf, '\n')) wrefresh(opwin);
	return(tmpln);
}

int addOpChrs(char *outputdata)
{
	if(!curses_on){
		printf("%s", outputdata);
		return(1);
	}
	waddnstr(opwin, outputdata, strlen(outputdata));
	wrefresh(opwin);
	return(1);
}

int addOpText(char *outputdata)
{
	if(!curses_on){
		printf("%s\n", outputdata);
		return(1);
	}
	waddnstr(opwin, outputdata, strlen(outputdata));
	waddch(opwin, '\n');
	wrefresh(opwin);
	return(1);
}

int addScrlText(char *outputdata, int oplines)
{char *temptx, *opstr, *nexstr; int txsize, rowsout=0;

	if(!curses_on){
		printf("%s", outputdata);
		return(1);
	}
	txsize=strlen(outputdata);
	temptx = (char *) malloc(txsize+1);
	if(!temptx) return(0); temptx[txsize]=0;
	memcpy(temptx, outputdata, txsize);
	opstr=nexstr=temptx;
	while(nexstr){
		opstr=strsep(&nexstr, "\n");
		waddnstr(opwin, opstr, strlen(opstr));
		waddch(opwin, '\n');
		rowsout++;
		if(rowsout>=oplines){
			waddstr(opwin, "Hit any key to continue\n");
			wrefresh(opwin);
			getch();
			rowsout=0;
		}
	}
	waddch(opwin, '\n');
	wrefresh(opwin);
	
	free(temptx);
	return(1);
}

bool check_str(char *cmdstr, char *teststr, int charct)
{char *strbg;
	strbg=cmdstr;
	while(*strbg && *strbg==' ') strbg++;
	if(!memcmp(strbg, teststr, charct)) return(TRUE);
	return(FALSE);
}


