/*
    asc2UKIsc Ascii to UKI Shellcode conversion program
    Copyright (C) 2010  NOP Security

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "scancode_map.h"
#include "USBKeyInject.h"
#include "ns_nc_prompt.h"
#include "help_data.h"

#define EEMEMSIZE		0x10000

#define ckst(nom) check_str(currentcmd, nom, sizeof(nom)-1)

int cvtToCtlCodes(unsigned char *ascii_buf, int ascii_sz, unsigned char *ctlcode_buf, int ctlcode_sz, unsigned char modifiers)
{unsigned char lastbyte=-1, newmod=modifiers, lastmod=modifiers, *curctlc=ctlcode_buf; int x;
	
	for(x=0; x<ascii_sz && curctlc<(ctlcode_buf+ctlcode_sz); x++){
		if(ascii_buf[x] == '~' && ascii_buf[++x] != '~' && ascii_buf[x] >= 0x30 && ascii_buf[x] <= 0x7A){
			newmod = modifiers;
			if(newmod != lastmod){
				if((curctlc+2)>=(ctlcode_buf+ctlcode_sz)) return(curctlc-ctlcode_buf);
				*curctlc++ = EECMD_STMODKEYS;
				*curctlc++ = newmod;
				lastmod = newmod;
			}
			if(lastbyte == AsciiToSpecialMap[ascii_buf[x]-0x30]){
				if((curctlc+1)>=(ctlcode_buf+ctlcode_sz)) return(curctlc-ctlcode_buf);
				*curctlc++ = 0;
				lastbyte=0;
			}
			*curctlc++ = AsciiToSpecialMap[ascii_buf[x]-0x30];
			lastbyte = AsciiToSpecialMap[ascii_buf[x]-0x30];
		}
		else {
			newmod = AsciiToScancodeMap[(ascii_buf[x]<<1)] ^ modifiers;
			if(newmod != lastmod){
				if((curctlc+2)>=(ctlcode_buf+ctlcode_sz)) return(curctlc-ctlcode_buf);
				*curctlc++ = EECMD_STMODKEYS;
				*curctlc++ = newmod;
				lastmod = newmod;
			}
			if(lastbyte == AsciiToScancodeMap[(ascii_buf[x]<<1)+1]){
				if((curctlc+1)>=(ctlcode_buf+ctlcode_sz)) return(curctlc-ctlcode_buf);
				*curctlc++ = 0;
				lastbyte=0;
			}
			*curctlc++ = AsciiToScancodeMap[(ascii_buf[x]<<1)+1];
			lastbyte = AsciiToScancodeMap[(ascii_buf[x]<<1)+1];
		}
	}
	return(curctlc-ctlcode_buf);
}

int getAsciiFromSC(unsigned char scchr, unsigned char mods)
{int tx;

	for(tx=0; tx<sizeof(AsciiToScancodeMap); tx+=2)
		if(AsciiToScancodeMap[tx+1] == scchr &&
			(((AsciiToScancodeMap[tx]!=0)&&((mods&0x22)!=0)) ||
			 ((AsciiToScancodeMap[tx]==0)&&((mods&0x22)==0)) ))
			return(tx/2);
	for(tx=0; tx<sizeof(AsciiToSpecialMap); tx++)
		if(AsciiToSpecialMap[tx] == scchr)
			return((tx+0x30)*-1);
	return(0);
}

#define		printoln(chrline)	addOpChrs(chrline)

int parseCtlCodes(unsigned char *scbuf, int scct)
{int retsz=0, tx, oi, tmpx; char databuf[512]; unsigned char lastch=0, mods=0;
	
	for(tx=0; tx<scct; tx++){

		switch(scbuf[tx]){
		case EECMD_END:
			printoln("<End script>\n");
			tx=scct;
			break;
		case EECMD_RESTART:
			printoln("<Restart from beginning>\n");
			tx=scct;
			break;
		case EECMD_NOP:
			printoln("<NOP>"); lastch=scbuf[tx];
			break;
		case EECMD_STRHEX16: lastch=scbuf[tx];
			tmpx=16;
			printoln("<");
			while(tmpx--){
				sprintf(databuf, "\\x%02X", scbuf[++tx]);
				printoln(databuf);
			}
			printoln(">\n");
			break;
		case EECMD_STRHEXS: lastch=scbuf[tx];
			tmpx=scbuf[++tx];
			if(tmpx > HEXEN_BASE){
				switch(tmpx){
				case HEXEN_STR:
					printoln("<Set hex output to string hex>");
					break;
				case HEXEN_VAR:
					printoln("<Set hex output to variable hex>");
					break;
				case HEXEN_RAW:
					printoln("<Set hex output to raw hex>");
					break;
				default:
					break;
				}
				break;
			}
			if(tmpx > 16) tmpx=16;
			printoln("<");
			while(tmpx--){
				sprintf(databuf, "\\x%02X", scbuf[++tx]);
				printoln(databuf);
			}
			printoln(">\n");
			break;
		case EECMD_STRHEX: lastch=scbuf[tx];
			sprintf(databuf, "<\\x%02X>", scbuf[++tx]);
			printoln(databuf);
			break;
		case EECMD_GOTO: lastch=scbuf[tx];
			++tx;
			sprintf(databuf, "<Goto instruction %04X.>", (scbuf[tx]<<8)|scbuf[tx+1]); ++tx;
			printoln(databuf);
			break;
		case EECMD_JUMP: lastch=scbuf[tx];
			++tx;
			sprintf(databuf, "<Jump %04X instructions.>", (scbuf[tx]<<8)|scbuf[tx+1]); ++tx;
			printoln(databuf);
			break;
		case EECMD_DELAYHR: lastch=scbuf[tx];
			sprintf(databuf, "<Delay ~%d hours.>", scbuf[++tx]);
			printoln(databuf);
			break;
		case EECMD_DELAYMIN: lastch=scbuf[tx];
			sprintf(databuf, "<Delay ~%d minutes.>", scbuf[++tx]);
			printoln(databuf);
			break;
		case EECMD_DELAYSEC: lastch=scbuf[tx];
			sprintf(databuf, "<Delay ~%d seconds.>", scbuf[++tx]);
			printoln(databuf);
			break;
		case EECMD_STMODKEYS:
			sprintf(databuf, "<Mod:%02X>", scbuf[++tx]);
			if(!lastch || lastch >= EECMDSTART || (scbuf[tx] && scbuf[tx] != 2)) printoln(databuf);
			mods=scbuf[tx]; lastch=scbuf[tx-1];
			break;
		case EECMD_DELAYQ: lastch=scbuf[tx];
			printoln("<Delay ~1/4 second.>");
			break;
		case EECMD_CONNECT: lastch=scbuf[tx];
			printoln("<Connect>");
			break;
		case EECMD_DISCONNECT: lastch=scbuf[tx];
			printoln("<Disconnect>");
			break;
		default:
			if(scbuf[tx]<EECMDSTART){
				if(!scbuf[tx] && lastch>=EECMDSTART) printoln("<key release>");
				if(scbuf[tx] != lastch){
					if(scbuf[tx]){
						oi=getAsciiFromSC(scbuf[tx], mods);
						if(oi=='\n') printoln("\n");
						else if(oi<0){
							oi*=-1;
							sprintf(databuf, "~%c", oi&0xFF);
							printoln(databuf);
						}
						else{
							databuf[0]=oi; databuf[1]=0;
							printoln(databuf);
						}
					}
				}
				lastch=scbuf[tx];
			}
			break;
		}

	}
	
	return(retsz);
}

void usage(char *execnm)
{
	printf("\nUsage: %s -c or -n\n", execnm);
	printf("-c : Curses\n");
	printf("-n : No Curses\n\n");
	printf("ASCII to UKI Scancode\n");
	printf("(c) 2010 NOP Security\n\n");

	return;
}

void display_ops(int curins, int cureeptr, unsigned char modifs)
{int pos=cmdpos-2;
	if(!curses_on){
		printf("Cur:%04X End:%04X Mod:%02X>", curins, cureeptr, modifs);
		return;
	}
	mvprintw(pos, 2, "Cur: %04X ", curins);
	mvprintw(pos, 14, "End: %04X ", cureeptr);
	mvprintw(pos, 26, "Mod: %02X ", modifs);
	wrefresh(opwin);
	return;
}

int main(int argc, char **argv)
{int optionr, cureeptr, curins, cmdslen, running=1, tmpx, filefd; char currentcmd[512], *tmpch;
 unsigned char eebuffer[EEMEMSIZE+8], tmpbuf[16]; unsigned char modifs=0, newmod;
 bool curse=TRUE;
	
	if(argc<2){usage(argv[0]); return(0);}
	while((optionr=getopt(argc, argv, "nch")) != -1){
		switch(optionr){
			case 'c':
				curse=TRUE;
				break;
			case 'n':
				curse=FALSE;
				break;
			default:		usage(argv[0]); return(0);
				break;
		}
	}
	
	initdisp(curse);
	
	curins=cureeptr=0;
	addOpText("Ascii to UKI scancode. v1.0\n(c) 2010 NOP Security\nType 'help' for more options.\n");
	display_ops(curins, cureeptr, modifs);
	memset(eebuffer, 0, EEMEMSIZE);
	
	while(running){
		cmdslen=parse_cmdln(currentcmd, 512);
		if(cmdslen){
			if(curses_on) addOpText(currentcmd);
			if(ckst("q\0")||ckst("x\0")||ckst("quit")||ckst("exit")) running=0;
			else if(ckst("?\0")||ckst("h\0")||ckst("help")) addScrlText(help_data, LINES-10);
			else if(ckst("list\0")) addScrlText(SpecialMapDescrip, LINES/2);
			else if(ckst("convert\0")){
				addOpText("Entering convert mode. Enter ~ on a line by itself to end, or ~ at the end of a line removes the trailing newline.\n");
				tmpx=curins;
				cmdslen=parse_cmdln(currentcmd, 512);
				while(!ckst("~\0")){
					if(curses_on) addOpText(currentcmd);
					if(currentcmd[cmdslen-1] != '~'){ strcat(currentcmd, "\n"); cmdslen++; }
					else currentcmd[cmdslen-1]=0;
					curins+=cvtToCtlCodes((unsigned char *)currentcmd, cmdslen, eebuffer+curins, EEMEMSIZE-curins, modifs);
					if(curses_on) display_ops(curins, cureeptr, modifs);
					cmdslen=parse_cmdln(currentcmd, 512);
				}
				if(curins<EEMEMSIZE && eebuffer[curins-1]) eebuffer[curins++]=0;
				addOpTextf("Added %d characters to buffer for text.\n", curins-tmpx);
			}
			else if(ckst("mod\0")) addScrlText(mod_help, LINES/2);
			else if(ckst("mod ")){
				tmpch=currentcmd; strsep(&tmpch, " ");
				newmod=strtol(tmpch, NULL, 0)&0xFF;
				if(modifs != newmod && curins+1 < EEMEMSIZE){
					modifs=newmod; eebuffer[curins++]=EECMD_STMODKEYS; eebuffer[curins++]=modifs;
					addOpTextf("Appending 2 bytes for modifier change to %02X\n", newmod);
				}
				else addOpTextf("No change made to modifier keys.\n");
			}
			else if(ckst("null\0")||ckst("0\0")){
				if(curins < EEMEMSIZE){ eebuffer[curins++]=0; addOpTextf("Appending 1 null to buffer.\n");}
			}
			else if(ckst("fin\0")){
				if(curins < EEMEMSIZE){ eebuffer[curins++]=EECMD_END; addOpTextf("Appending finish command to buffer.\n");}
			}
			else if(ckst("fin ")){
				tmpch=currentcmd; strsep(&tmpch, " ");
				tmpx=strtol(tmpch, NULL, 0);
				addOpTextf("Appending %d finish commands to buffer.\n", tmpx);
				while(tmpx-- > 0) if(curins < EEMEMSIZE) eebuffer[curins++]=EECMD_END;
			}
			else if(ckst("restart\0")){
				if(curins < EEMEMSIZE){ eebuffer[curins++]=EECMD_RESTART; addOpTextf("Appending restart command to buffer.\n");}
			}
			else if(ckst("nop\0")){
				if(curins < EEMEMSIZE){ eebuffer[curins++]=EECMD_NOP; addOpTextf("Appending NOP to buffer.\n");}
			}
			else if(ckst("qdelay\0")){
				if(curins < EEMEMSIZE){ eebuffer[curins++]=EECMD_DELAYQ; addOpTextf("Appending 1 Q delay to buffer.\n");}
			}
			else if(ckst("connect\0")){
				if(curins < EEMEMSIZE){ eebuffer[curins++]=EECMD_CONNECT; addOpTextf("Appending connect to buffer.\n");}
			}
			else if(ckst("disconnect\0")){
				if(curins < EEMEMSIZE){ eebuffer[curins++]=EECMD_CONNECT; addOpTextf("Appending disconnect to buffer.\n");}
			}
			else if(ckst("gotoins ")){
				tmpch=currentcmd; strsep(&tmpch, " ");
				tmpx=strtol(tmpch, NULL, 0)&0xFFFF;
				if(tmpx && curins+2 < EEMEMSIZE){
					eebuffer[curins++]=EECMD_GOTO;
					eebuffer[curins++]=(tmpx>>8)&0xFF;
					eebuffer[curins++]=tmpx&0xFF;
					addOpTextf("Adding goto instruction address %04X.\n", tmpx);
				}
				else addOpTextf("Delay must be a possible delay time.\n");
			}
			else if(ckst("jump ")){
				tmpch=currentcmd; strsep(&tmpch, " ");
				tmpx=strtol(tmpch, NULL, 0)&0xFFFF;
				if(tmpx && curins+2 < EEMEMSIZE){
					eebuffer[curins++]=EECMD_JUMP;
					eebuffer[curins++]=(tmpx>>8)&0xFF;
					eebuffer[curins++]=tmpx&0xFF;
					addOpTextf("Adding jump of %04X instructions.\n", tmpx);
				}
				else addOpTextf("Delay must be a possible delay time.\n");
			}
			else if(ckst("secdelay ")){
				tmpch=currentcmd; strsep(&tmpch, " ");
				tmpx=strtol(tmpch, NULL, 0)&0xFF;
				if(tmpx>0 && curins+1 < EEMEMSIZE){
					eebuffer[curins++]=EECMD_DELAYSEC; eebuffer[curins++]=tmpx&0xFF;
					addOpTextf("Adding %d second delay to buffer.\n", tmpx);
				}
				else addOpTextf("Delay must be a possible delay time.\n");
			}
			else if(ckst("mindelay ")){
				tmpch=currentcmd; strsep(&tmpch, " ");
				tmpx=strtol(tmpch, NULL, 0)&0xFF;
				if(tmpx>0 && curins+1 < EEMEMSIZE){
					eebuffer[curins++]=EECMD_DELAYMIN; eebuffer[curins++]=tmpx&0xFF;
					addOpTextf("Adding %d minute delay to buffer.\n", tmpx);
				}
				else addOpTextf("Delay must be a possible delay time.\n");
			}
			else if(ckst("hourdelay ")){
				tmpch=currentcmd; strsep(&tmpch, " ");
				tmpx=strtol(tmpch, NULL, 0)&0xFF;
				if(tmpx>0 && curins+1 < EEMEMSIZE){
					eebuffer[curins++]=EECMD_DELAYHR; eebuffer[curins++]=tmpx&0xFF;
					addOpTextf("Adding %d hour delay to buffer.\n", tmpx);
				}
				else addOpTextf("Delay must be a possible delay time.\n");
			}
			else if(ckst("clear\0")){
				memset(eebuffer, 0, EEMEMSIZE);
				curins=cureeptr=modifs=0;
				addOpText("Memory cleared and reset.\n");
			}
			else if(ckst("d\0")||ckst("dump")){
				for(tmpx=0; tmpx<EEMEMSIZE && tmpx<cureeptr; tmpx++){
					if((tmpx&0xF) == 0) addOpTextf("%04X", tmpx&0xFFF0);
					if(tmpx == curins) addOpTextf("*"); else addOpTextf(" ");
					addOpTextf("%02X", eebuffer[tmpx]);
					if((tmpx&0xF) == 0xF) addOpTextf("\n");
				}
				addOpTextf("\n");
			}
			else if(ckst("parse\0")){
				parseCtlCodes(eebuffer, cureeptr);
				addOpTextf("Parse complete\n");
			}
			else if(ckst("end ")){
				tmpch=currentcmd; strsep(&tmpch, " ");
				cureeptr=strtol(tmpch, NULL, 0);
				if(cureeptr > EEMEMSIZE) cureeptr=EEMEMSIZE;
				if(curins > cureeptr) curins=cureeptr;
				addOpTextf("End point set to %04X\n", cureeptr);
			}
			else if(ckst("g ")||ckst("goto ")){
				tmpch=currentcmd; strsep(&tmpch, " ");
				curins=strtol(tmpch, NULL, 0);
				addOpTextf("Current edit point set to %04X\n", curins);
			}
			else if(ckst("save ")){
				tmpch=currentcmd; strsep(&tmpch, " ");
				filefd=creat(tmpch, 0644);
				if(filefd!=-1){
					tmpx=write(filefd, eebuffer, cureeptr);
					addOpTextf("%d bytes written to %s\n", tmpx, tmpch);
					close(filefd);
				}
				else addOpTextf("File create failed.\n");
			}
			else if(ckst("read ")){
				tmpch=currentcmd; strsep(&tmpch, " ");
				filefd=open(tmpch, O_RDONLY);
				if(filefd!=-1){
					tmpx=read(filefd, eebuffer+curins, EEMEMSIZE-curins);
					addOpTextf("Read %d bytes from %s\n", tmpx, tmpch);
					curins+=tmpx;
					close(filefd);
				}
				else addOpTextf("File open failed.\n");
			}
			else if(ckst("encode ")){
				tmpch=currentcmd; strsep(&tmpch, " ");
				filefd=open(tmpch, O_RDONLY);
				if(filefd!=-1){
					do{
					tmpx=read(filefd, tmpbuf, 16);
					if(tmpx==16 && curins+17<EEMEMSIZE){
						eebuffer[curins++]=EECMD_STRHEX16;
						memcpy(eebuffer+curins, tmpbuf, 16);
						curins+=tmpx;
						}
					else if(tmpx==1 && curins+2<EEMEMSIZE){
						eebuffer[curins++]=EECMD_STRHEX;
						eebuffer[curins++]=tmpbuf[0];
						}
					else if(tmpx>0 && curins+tmpx+2<EEMEMSIZE){
						eebuffer[curins++]=EECMD_STRHEXS;
						eebuffer[curins++]=tmpx&0x1F;
						memcpy(eebuffer+curins, tmpbuf, tmpx);
						curins+=tmpx;
						}
					} while(tmpx>0);
					close(filefd);
					addOpTextf("Encoded file %s to hex.\n", tmpch);
					
				}
				else addOpTextf("File open failed.\n");
			}
			else if(!curses_on) system(currentcmd);
		}
		if(curins > EEMEMSIZE) curins=EEMEMSIZE;
		if(curins > cureeptr) cureeptr=curins;
		display_ops(curins, cureeptr, modifs);
	}
	
	stopdisp();
	printf("\n");
	return(0);
}


